// SPDX-License-Identifier: GPL-2.0-only
//
// events.*.*.log parser and summarizer script
//
// Copyright (c) 2021 Vitaly Chikunov <vt@altlinux.org>
//

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <set>
#include <string>
#include <string_view>
#include <unordered_map>
#include <unordered_set>
#include <unistd.h>
#include <re2/re2.h>
#include <absl/strings/string_view.h>

using namespace std;
using namespace re2;

time_t time2sec(const string &s)
{
    istringstream ss(s);
    tm t{};
    ss >> get_time(&t, "%Y-%b-%d %H:%M:%S");
    return mktime(&t);
}

string sec2time(time_t s)
{
    int hh = s / 3600;
    int mm = s % 3600 / 60;
    int ss = s % 60;
    char buf[32];
    snprintf(buf, sizeof(buf), "%d:%02d:%02d", hh, mm, ss);
    size_t n = strspn(buf, "0:");
    if (n > 0 && buf[n] == '\0')
	n--;
    return string(buf + n);
}

bool parse_name(const string& in, string& git, string& tag)
{
    size_t j = in.find_first_of(' ');
    size_t n;
    if (j == string::npos) { // .src.rpm
	static const string _src_rpm(".src.rpm");
	if (!in.ends_with(_src_rpm))
	    return false;
	n = in.rfind('-');
	if (n == 0 || n == string::npos)
	    return false;
	n = in.rfind('-', n - 1);
	if (n == 0 || n == string::npos)
	    return false;
	git = in.substr(0, n++);
	tag = in.substr(n, in.size() - n - _src_rpm.size());
    } else { // .git
	static const string _git(".git");
	string_view repo(in.c_str(), j);
	if (!repo.ends_with(_git))
	    return false;
	repo.remove_suffix(_git.size());
	git = repo;
	tag = in.substr(j + 1);

	// Special case.
	if (git == "kernel-modules") {
	    static const string kernel_modules_("kernel-modules-");
	    j = tag.find(kernel_modules_);
	    if (j == string::npos)
		return true;
	    size_t k = j + kernel_modules_.size();
	    n = tag.rfind('-');
	    if (n <= k || n == string::npos)
		return true;
	    n = tag.rfind('-', n - 1);
	    if (n <= k || n == string::npos)
		return true;
	    git = '@' + tag.substr(j, n - j);
	}
    }
    return true;
}

int _stoi(const string& str)
{
    try {
	return stoi(str);
    } catch (logic_error& e) {
	return 0;
    }
}

void print_report(istream& is)
{
    unordered_set<string> nnrs; // No Need to Rebuild-s skip list
    unordered_map<string, string> starts; // date_time
    unordered_map<string, string> stops;  // state
    set<int> subtasks;
    unordered_set<int> dels; // removed subtasks
    set<string> arches;
    unordered_map<int, string> names;
    unordered_map<string, size_t> cols; // width of column

// 2020-Jul-22 13:01:21 :: task #255239 for sisyphus started by vt:
// 2020-Jul-10 11:52:14 :: [ppc64le] #100 kernel-image-ovz-el7.git 3.10.0-alt3.1127.10.1.vz7.162.5: build start
// 2020-Jul-10 11:52:51 :: [x86_64] kernel-image-ovz-el7.git 3.10.0-alt3.1127.10.1.vz7.162.5: remote: no need to rebuild
// 2020-Jul-10 11:52:38 :: [i586] #100 kernel-image-ovz-el7.git 3.10.0-alt3.1127.10.1.vz7.162.5: build SKIPPED
// 2020-Jul-10 11:52:54 :: [x86_64] #100 kernel-image-ovz-el7.git 3.10.0-alt3.1127.10.1.vz7.162.5: build OK
// 2020-Jul-23 14:07:01 :: [i586] #100 perl-DBIx-SearchBuilder-1.68-alt1.src.rpm: build start
// 2020-Jul-23 14:44:34 :: [x86_64] perl-DBIx-SearchBuilder-1.68-alt1.src.rpm: remote: no need to rebuild

    string line;
    while (getline(is, line)) {
	absl::string_view input(line);
	string subtask;
	static RE2 re_removed("#(\\d+) removed");
	if (input[0] == '#' && RE2::FullMatch(input, re_removed, &subtask)) {
	    dels.insert(_stoi(subtask));
	    continue;
	}
	string date_time;
	static RE2 re_date_time("(2\\d{3}-\\w{3}-\\d\\d \\d\\d:\\d\\d:\\d\\d) :: ");
	if (!RE2::Consume(&input, re_date_time, &date_time))
	    continue;
	static string test_only("test-only ");
	if (input.starts_with(test_only))
	    input.remove_prefix(test_only.size());
	static RE2 re_task_started("^task #\\d+ for \\w+ \\w+ by");
	if (RE2::PartialMatch(input, re_task_started)) {
	    nnrs.clear();
	    continue;
	}
	string arch;
	static RE2 re_arch("\\[(\\w+)\\] ");
	if (!(input[0] == '[' && RE2::Consume(&input, re_arch, &arch)))
	    continue;
	static RE2 re_subtask("^#(\\d+) ");
	if (input[0] == '#')
	    RE2::Consume(&input, re_subtask, &subtask); // optional
	static string remote_nnr(": remote: no need to rebuild");
	if (input.ends_with(remote_nnr)) {
	    input.remove_suffix(remote_nnr.size());
	    nnrs.insert(arch + ' ' + string(input));
	    continue;
	}
	if (subtask.empty())
	    continue;
	// Only match 'build' lines.
	string state;
	string nnr; // key to nnrs set
	static RE2 re_build("([^:]+): build (\\w+)\\s*");
	if (!RE2::Consume(&input, re_build, &nnr, &state))
	    continue;
	int n = _stoi(subtask);
	// Now 'input' contains text after build state.
	string git; // %name
	string tag; // %version-%release
	if (!parse_name(nnr, git, tag))
	    continue;
	nnr = arch + ' ' + nnr;
	arches.insert(arch);

	string n_a = subtask + ' ' + arch; // key to starts/stops sets
	string start, stop;
	if (state == "start") {
	    starts[n_a] = date_time;
	    continue;
	} else if (input == "(cached)" || state == "swiftly") {
	    // 'OK (cached)' and 'swiftly SKIPPED' work similar to NNR
	    stop = stops[n_a];
	    if (stop.empty())
		state = "-";
	    else
		state = stop;
	} else if (state == "OK") {
	    // If 'build OK' ...
	    string start = starts[n_a];
	    string stop  = stops[n_a];
	    if (nnrs.count(nnr)) {
		if (stop.empty())
		    state = "-";
		else
		    state = stop;
	    } else {
		// Normal 'build OK' - calc elapsed time.
		if (!start.empty()) {
		    time_t ts1 = time2sec(start);
		    time_t ts2 = time2sec(date_time);
		    state = sec2time(ts2 - ts1);
		} else
		    state = "-";
	    }
	} else if (state == "SKIPPED") {
	    state = "-";
	} else {
	    transform(state.begin(), state.end(), state.begin(), ::tolower);
	}
	stops[n_a] = state;
	if (state == "-")
	    continue;
	subtasks.insert(n);
	names[n] = git;
	cols[arch] = max(state.size(), cols[arch]);
    }

    if (subtasks.empty() || arches.empty())
	return;

    // Account width of headers.
    size_t cols_subtask = strlen("subtask");
    for (auto& n : subtasks) {
	string subtask = '#' + to_string(n);
	cols_subtask = max(subtask.size(), cols_subtask);
    }
    size_t cols_name = strlen("name");
    for (auto& n : names)
	cols_name = max(n.second.size(), cols_name);
    for (auto& a : arches)
	cols[a] = max(a.size(), cols[a]);

    // And now...
    int tty = isatty(1);
    printf("%-*s", (int)cols_subtask, "subtask");
    printf("  %-*s", (int)cols_name, "name");
    for (auto& a : arches)
	printf("  %*s", (int)cols[a], a.c_str());
    printf("\n");

    for (auto& n : subtasks) {
	if (tty && dels.count(n))
	    printf("\e[2m");
	string git = names[n];
	string subtask = '#' + to_string(n);
	printf("%*s", (int)cols_subtask, subtask.c_str());
	printf("  %-*s", (int)cols_name, git.c_str());
	for (auto& a : arches) {
	    string n_a = to_string(n) + ' ' + a;
	    string state = stops[n_a];
	    if (state.empty())
		state = "?";
	    printf("  %*s", (int)cols[a], state.c_str());
	}
	if (dels.count(n)) {
	    printf(" R");
	    if (tty)
		printf("\e[m");
	}
	printf("\n");
    }
}

int main(int argc, char **argv)
{
    ios_base::sync_with_stdio(false); // speed up getline

    if (argc > 1)
	for (int i = 1; i < argc; i++) {
	    ifstream is(argv[i]);
	    print_report(is);
	}
    else {
	if (isatty(0))
	    cerr << "Waiting for data from stdin.\n";
	print_report(cin);
    }

    return 0;
}
